#include <windows.h>
#include <stdio.h>
#include <memory.h>

#include "type.h"
#include "aspi.h"
#include "wnaspi32.h"
#include "scsidefs.h"
#include "resource.h"

static CdrConfS cdrConfig;

static volatile UINT8 aspiWorking = 0;
static volatile UINT8 aspiError   = 0;

static SRB_ExecSCSICmd myExeSrb;
static InquiryS        myInq;

static void ASPI32Post (SRB_ExecSCSICmd *Srb)
{
    if(Srb->SRB_Status != SS_COMP && Srb->SRB_Status != SS_NO_DEVICE)
    {
        aspiError = Srb->SRB_Status;
    }
    else aspiError = 0;

    aspiWorking = 0;
}

static int SendInquiry (int adapter_id, int target_id)
{
    memset(&myExeSrb,0,sizeof(myExeSrb));
    memset(&myInq,0,sizeof(myInq));
    aspiWorking = 1;

    myExeSrb.SRB_Cmd        = SC_EXEC_SCSI_CMD;
    myExeSrb.SRB_HaId       = adapter_id;
    myExeSrb.SRB_Flags      = SRB_DIR_IN | SRB_POSTING;
    myExeSrb.SRB_Target     = target_id;
    myExeSrb.SRB_BufPointer = (unsigned char *)&myInq;
    myExeSrb.SRB_BufLen     = INQUIRY_BUFFER_LENGTH;
    myExeSrb.SRB_SenseLen   = SENSE_LEN;
    myExeSrb.SRB_CDBLen     = 6;
    myExeSrb.CDBByte [0]    = SCSI_INQUIRY;
    myExeSrb.CDBByte [4]    = INQUIRY_BUFFER_LENGTH;
    myExeSrb.SRB_PostProc   = (void *) ASPI32Post;

    return SendASPI32Command(&myExeSrb);
}

static int SCSIinquiry(int adapter_id, int target_id, char * buffer)
{
    int i;

    SendInquiry(adapter_id,target_id);
    while (aspiWorking);
    for (i=0;i<24;i++) buffer[i] = myInq.Vendor[i];
    buffer[24]=0;
    if (myInq.devType == DTYPE_CROM || myInq.devType == DTYPE_WORM)
        return	1;
    else
        return 0;
}

int EnumerateCD(HWND hwnd)
{
    DWORD dwSupportInfo;
    int   status, hostcount, adapterid, targetid;
    char  tmpbuf[256];

    dwSupportInfo = GetASPI32SupportInfo();

    status = (dwSupportInfo>>8)&255;
    hostcount = dwSupportInfo&255;

    if (hostcount ==0 || status != SS_COMP) {
        MessageBox(hwnd,"ASPI init fail","System Error",MB_OK);
        return FPSE_ERR;
    }

    memset(&cdrConfig,0,sizeof(cdrConfig));

    for(adapterid=0;adapterid<hostcount;adapterid++)
        for(targetid=0;targetid<MAX_ID;targetid++)
            if (SCSIinquiry(adapterid,targetid,tmpbuf)) {

                strcpy(cdrConfig.cdromsA[cdrConfig.cdromsC],tmpbuf);
                cdrConfig.cdrCa[cdrConfig.cdromsC] = adapterid;
                cdrConfig.cdrCt[cdrConfig.cdromsC] = targetid;

                cdrConfig.cdromsC++;
            }

    if (!cdrConfig.cdromsC) return FPSE_ERR;

    return FPSE_OK;
}

int GetCode(int pos)
{
    return (UINT32)(cdrConfig.cdrCa[pos]) |
           ((UINT32)(cdrConfig.cdrCt[pos])<<8);
}

void SetToCombo(HWND hwnd)
{
    int x;
    char buf[256];

    for (x=0;x<cdrConfig.cdromsC;x++)
    {
        sprintf(buf,"%d:%d:%d %s",
                x,cdrConfig.cdrCa[x],cdrConfig.cdrCt[x],
                cdrConfig.cdromsA[x]);

        SendMessage((HWND)GetDlgItem(hwnd,IDC_COMBODRIVES),
                    CB_ADDSTRING,0,(LONG)buf);
    }
    SendMessage((HWND)GetDlgItem(hwnd,IDC_COMBODRIVES),
                CB_SETCURSEL,0,0);

}
