#include <windows.h>
#include <stdio.h>
#include <memory.h>

#include "type.h"
#include "Cd.h"
#include "Win32Def.h"
#include "wnaspi32.h"
#include "scsidefs.h"

// #define DEBUG_ASPI

typedef struct {
	UINT8 DevType;
	char  reserved[7];
	char  Vendor[24];
} Inquiery;

typedef struct {
	unsigned char reserved;
	unsigned char adr;
	unsigned char trackno;
	unsigned char reserved1;
	unsigned char addr[4];
} TOCENTRY;

typedef struct {
	unsigned char size[2];
	unsigned char first;
	unsigned char last;
	TOCENTRY track[100];
} TOC;

typedef	struct {
	UINT8 minute;
	UINT8 second;
	UINT8 frame;
} LOC;

#define	INT2BCD(a)	(((a)/10)*16+((a)%10))
#define	BCD2INT(a)	(((a)>>4)*10+((a)&0xf))

#define SECTOR_SIZE 2352

static int   adapter_id = 0;
static int   target_id = 0;
static int   target_lun = 0;
static UINT8 readbuf[SECTOR_SIZE];

#define	MAX_ID	8

static int  loc2int(LOC *loc);
static void int2loc(LOC *loc,int n);
static int  GetDeviceType(int adapterid,int targetid);
static void ExecReadTOC(void *buf,int size,int first);
static void ExecRead(unsigned long frame,void *buf,int bufsize);
static void ExecPlay(unsigned char *start,unsigned char *end);
static void ExecStop(void);
static void ExecModeSelect(int mode);

int CD_Open(UINT32 *par)
{
    DWORD dwSupportInfo;
    int status, hostcount;
#ifdef DEBUG_ASPI
    int adapterid, targetid;
#else
    FPSEWin32  *g = (FPSEWin32 *)par;
    char        buf[4];
#endif

    dwSupportInfo = GetASPI32SupportInfo();

    status = (dwSupportInfo>>8)&255;
    hostcount = dwSupportInfo&255;

    if (hostcount ==0 || status != SS_COMP) {
        printf("ASPI init fail %x",(int)dwSupportInfo);
        return FPSE_ERR;
    }

#ifdef DEBUG_ASPI
    for(adapterid=0;adapterid<hostcount;adapterid++) {
        for(targetid=0;targetid<MAX_ID;targetid++) {
            int type = GetDeviceType(adapterid,targetid);

            if (type == DTYPE_CDROM) {
                printf("adapter=%d target=%d\n",adapterid,targetid);
                adapter_id = adapterid;
                target_id = targetid;
                ExecModeSelect(0);
                CD_Wait();
                return FPSE_OK;
            }
        }
    }
#else
    g->ReadCfg("CDASPI","AdapterID",&buf);
    adapter_id = buf[0] - '0';
    g->ReadCfg("CDASPI","TargetID",&buf);
    target_id = buf[0] - '0';
    if (GetDeviceType(adapter_id,target_id) == DTYPE_CDROM) {
        printf("adapter=%d target=%d\n",adapter_id,target_id);
        ExecModeSelect(3);
        CD_Wait();
        return FPSE_OK;
    }
#endif
// not found cdrom
    return FPSE_ERR;
}

void CD_Close(void)
{
}

static TOC toc;
static struct {
	unsigned char size[2];
	unsigned char first,last;
	TOCENTRY track[1];
} readout;


int CD_GetTN(UINT8 *result)
{
    ExecReadTOC(&toc,sizeof(toc),0);
/*
	{ int i;
	for(i=0;i<=toc.last;i++) dump(&toc.track[i],8);
	}
*/
    result[1] = toc.first;
    result[2] = toc.last;

    return 0;
}

int CD_GetTD(UINT8 *result,int track)
{
    int n;

    ExecReadTOC(&toc,sizeof(toc),0);
    n = track-toc.first;

    result[1] = toc.track[n].addr[1]; /* minute */
    result[2] = toc.track[n].addr[2]; /* second */
    result[3] = toc.track[n].addr[3]; /* frame */

    return 0;
}

int CD_Play(UINT8 *param)
{
/* serch next track */
    ExecPlay(param,&readout.track[0].addr[2]);
    return 0;
}

int CD_Stop(void)
{
/* serch next track */
    ExecStop();
    return 0;
}

UINT8 *CD_Read(UINT8 *param)
{
    ExecRead(loc2int((LOC *)param),readbuf+12,SECTOR_SIZE);

// emulate header
    readbuf[0] = INT2BCD(param[0]);
    readbuf[1] = INT2BCD(param[1]);
    readbuf[2] = INT2BCD(param[2]);

    return readbuf;
}

void CD_GetSeek(UINT8 *par)
{

}

static  SRB_ExecSCSICmd     srb;
static  HANDLE              hEventSRB;
static  int                 LastStatus = SS_COMP;

static int ExecSCSICmd_sub(UINT8 *cdb,int cdb_size,UINT8 *out_buff,int out_size,BYTE flag)
{
        if (hEventSRB) CloseHandle(hEventSRB);

	hEventSRB = CreateEvent(NULL,TRUE,FALSE,NULL);
	if (hEventSRB==NULL) {
		printf("can't create event\n");
	}

	memset(&srb,0,sizeof(SRB_ExecSCSICmd));

	srb.SRB_Cmd = SC_EXEC_SCSI_CMD;
	srb.SRB_HaId = adapter_id;
	srb.SRB_Flags = flag | SRB_EVENT_NOTIFY;
	srb.SRB_Target = target_id;
//	srb.SRB_Lun = target_lun;
	srb.SRB_SenseLen = SENSE_LEN;
	srb.SRB_PostProc = (LPVOID)hEventSRB;

	srb.SRB_BufPointer = out_buff;
	srb.SRB_BufLen = out_size;
	srb.SRB_CDBLen = cdb_size;
	memcpy(srb.CDBByte,cdb,cdb_size);

	LastStatus = SendASPI32Command(&srb);
//	printf("status:%x \n",status);
    return 0;
}

int CD_Wait()
{
    if (LastStatus == SS_PENDING) {
        WaitForSingleObject(hEventSRB,INFINITE);
//        printf("wait\n");

        CloseHandle(hEventSRB);
        LastStatus = SS_COMP;

        if (srb.SRB_Status != SS_COMP) {
//            printf("%x:error %x\n",cdb[0],srb.SRB_Status);
//            return FPSE_ERR;
        }
    }
    return FPSE_OK;
}

static int ExecSCSICmd(UINT8 *cdb,int cdb_size,UINT8 *out_buff,int out_size)
{
    return ExecSCSICmd_sub(cdb,cdb_size,out_buff,out_size,SRB_DIR_IN);
}

static int loc2int(LOC *loc)
{
    return ( (loc->minute)*60 + (loc->second) -2 )*75 + (loc->frame);
}

static void int2loc(LOC *loc,int n)
{
	int minute = n/75+2;
	loc->frame = INT2BCD(n%75);
	loc->minute = INT2BCD(minute/60);
	loc->second = INT2BCD(minute%60);
}

static int GetDeviceType(int adapterid,int targetid)
{
	SRB_GDEVBlock srbGDEVBlock;
	memset(&srbGDEVBlock,0,sizeof(srbGDEVBlock));
	srbGDEVBlock.SRB_Cmd = SC_GET_DEV_TYPE;
	srbGDEVBlock.SRB_HaId = adapterid;
	srbGDEVBlock.SRB_Target = targetid;
	SendASPI32Command((LPSRB)&srbGDEVBlock);
	if (srbGDEVBlock.SRB_Status != SS_COMP) return -1;
	return srbGDEVBlock.SRB_DeviceType;
}

static void ExecReadTOC(void *buf,int size,int first)
{
	UINT8 cdb[10];

	memset(cdb,0,sizeof(cdb));

	cdb[0] = SCSI_READ_TOC;
	cdb[1] = 2; /* MSF formst */
	cdb[6] = first;
	cdb[7] = size>>8;
	cdb[8] = size&255;

	ExecSCSICmd(cdb,sizeof(cdb),buf,size);
}

static void ExecRead(unsigned long frame,void *buf,int bufsize)
{
	UINT8 cdb[10];

	memset(cdb,0,sizeof(cdb));

	cdb[0] = SCSI_READ10;
	cdb[2] = frame>>24;
	cdb[3] = frame>>16;
	cdb[4] = frame>>8;
	cdb[5] = frame;
	cdb[7] = 0;
	cdb[8] = 1; /* read 1 frames */

	ExecSCSICmd(cdb,sizeof(cdb),buf,bufsize);
}

static void ExecPlay(unsigned char *start,unsigned char *end)
{
	UINT8 cdb[10];

	memset(cdb,0,sizeof(cdb));

	cdb[0] = SCSI_PLAYAUDMSF;
	cdb[3] = start[0];
	cdb[4] = start[1];
	cdb[5] = start[2];
	cdb[6] = end[0];
	cdb[7] = end[1];
	cdb[8] = end[2];

	ExecSCSICmd(cdb,sizeof(cdb),NULL,0);
}

static void ExecStop(void)
{
	UINT8 cdb[10];

	memset(cdb,0,sizeof(cdb));

	cdb[0] = 0x4B; /* Pause command */
	cdb[8] = 0; /* 0:pause 1:resume */

	ExecSCSICmd(cdb,sizeof(cdb),NULL,0);
}

static void ExecModeSelect(int mode)
{
	UINT8 cdb[6],param[12+4];

	memset(cdb,0,sizeof(cdb));
	memset(param,0,sizeof(param));

	cdb[0] = SCSI_MODE_SEL6;
	cdb[1] = 0x10; /* SCSI-2 Page format */
	cdb[4] = 12; //sizeof(param);

	param[3] = 8; /* size of block descpritor */
	param[4] = mode; /* density code 0:default 1:2048 2:2336 3:2340 4:audio */
	if (mode==0 || mode==1) {
		param[10] = (2048>>8) & 0xFF;
		param[11] = (2048) & 0xFF;
	} else {
		param[10] = (SECTOR_SIZE>>8) & 0xFF;
		param[11] = (SECTOR_SIZE) & 0xFF;
	}

	ExecSCSICmd_sub(cdb,sizeof(cdb),param,12 /*sizeof(param)*/,SRB_DIR_OUT);
}

#ifdef DEBUG_ASPI
static void dump(UINT8 *data, int len)
{
    int   i,j;
    UINT8 c;

    for(i=0;i<len;i++)
    {
        if ((i&15)==0) printf("%04x:",i & ~15);
        printf("%02x ",data[i]);
        if ((i&15)==15) {
            for (j=0; j<16; j++)
            {
                c = data[(i & ~15)+j];
                if (c > 31 && c < 128) printf("%c",c);
                                  else printf(".");
            }
            printf("\n");
        }
    }

    printf("\n");
}

int main(int argc,char **argv)
{
    UINT8 param[8];
    int   i,res;
    char *buf;

    printf("FPSE CDASPI TEST\n\nInit...\n");
    res = CD_Open(NULL);
    if (res != FPSE_OK) return 1;
    printf("Selecting default mode...\n");
    ExecModeSelect(0);
    printf("Start reading...\n\n");

    for (i=0; i<23; i++)
    {
        param[0] = 0;
        param[1] = 2;
        param[2] = i;

        memset(readbuf,0,sizeof(readbuf));

        buf = CD_Read(param);
        res = CD_Wait();
        if (res == FPSE_OK)
        {
            printf("frame %d:\n",i);
            dump(buf,SECTOR_SIZE);
        } else {
            printf("frame %d - read error\n",i);
        }
    }

    CD_Close();

    return 0;
}
#endif
